////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2018 Josef Spidlen, Ph.D., FlowJo, LLC
//
// License
// The software is distributed under the terms of the 
// Artistic License 2.0
// http://www.r-project.org/Licenses/Artistic-2.0
// 
// Disclaimer
// This software and documentation come with no warranties of any kind.
// This software is provided "as is" and any express or implied 
// warranties, including, but not limited to, the implied warranties of
// merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the  copyright holder be liable for any direct, 
// indirect, incidental, special, exemplary, or consequential damages
// (including but not limited to, procurement of substitute goods or 
// services; loss of use, data or profits; or business interruption)
// however caused and on any theory of liability, whether in contract,
// strict liability, or tort arising in any way out of the use of this 
// software.    
//////////////////////////////////////////////////////////////////////////////

package com.flowjo.plugin.FitSNE;
import java.awt.Component;
import java.awt.Dimension;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.swing.Box;
import javax.swing.DefaultListModel;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JCheckBox;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.SwingConstants;

import com.flowjo.plugin.FitSNE.utils.FilenameUtils;
import com.flowjo.plugin.FitSNE.utils.DoubleNumberFormatter;
import com.treestar.flowjo.application.workspace.manager.FJApplication;
import com.treestar.flowjo.core.nodes.PopNode;
import com.treestar.flowjo.core.CSVSample;
import com.treestar.flowjo.engine.utility.RFlowCalculator;
import com.treestar.flowjo.engine.utility.R_Algorithm;
import com.treestar.lib.FJPluginHelper;
import com.treestar.lib.PluginHelper;
import com.treestar.lib.core.ExportFileTypes;
import com.treestar.lib.core.ExternalAlgorithmResults;
import com.treestar.lib.data.StringUtil;
import com.treestar.lib.file.FJFileRef;
import com.treestar.lib.file.FJFileRefFactory;
import com.treestar.lib.gui.FJList;
import com.treestar.lib.gui.GuiFactory;
import com.treestar.lib.gui.HBox;
import com.treestar.lib.gui.numberfields.FJFormattedTextField;
import com.treestar.lib.gui.numberfields.RangedDoubleTextField;
import com.treestar.lib.gui.numberfields.RangedIntegerTextField;
import com.treestar.lib.gui.panels.FJLabel;
import com.treestar.lib.gui.swing.FJCheckBox;
import com.treestar.lib.gui.swing.FJComboBox;
import com.treestar.lib.xml.SElement;

public class FItSNE extends R_Algorithm {
	
	public static final String pluginVersion = "1.3";
	public static final String pluginName = "FItSNE";
    public static final String cellIdParName = "CellId";
    public static final String Zero = "0";
    public static final String One = "1";
    public static final String True = "true";
    public static final String False = "false";

    public static final String citingLabel   = "&nbsp;<br/><b>If you are using this plugin, please cite:</b><br/>George C. Linderman, Manas Rachh, Jeremy G. Hoskins, Stefan Steinerberger, Yuval Kluger. Efficient Algorithms for t-distributed Stochastic Neighborhood Embedding. https://arxiv.org/abs/1712.09005";
    
    // Required binaries for OSX and Win, respectively. The first one needs to be the executable one.
    public static final String[] OSXResources = { "binaries/fast_tsne" };
    public static final String[] WinResources = { "binaries/FItSNE.exe", "binaries/libfftw3-3.dll", "binaries/libfftw3f-3.dll", "binaries/libfftw3l-3.dll" };
    
    public static final String sIconName = "src/images/FIt-SNE_Icon16.png";
    public static Icon myIcon = null;
    public static final String Failed = "Failed";
    public static final String alreadyExists = "AlreadyExists";
	
	public static final int labelW = 628;
	public static final int labelH = 25;
	
	public static final String keepResultsLabel   = "Save the R script and output messages";
	public static final String keepResultsToolTip = "Keep a file that shows execution of the script.";
    
    private FJComboBox fDimsCombo = new FJComboBox(dimOptionsHuman);
    public static final String dimsComboText = "t-SNE dimensions";
    public static final String dimsComboTooltip = "Number of dimensions in the t-SNE output; you can create 1 or 2 derived parameters.";
    public static final String dimOptionName = "dim";
    public static final String[] dimOptionsHuman = new String[] { "1", "2" };
    public static final String[] dimOptions = new String[] { "1", "2" };

    private FJComboBox fAlgCombo = new FJComboBox(algOptionsHuman);
    public static final String algComboText = "Implementation";
    public static final String algComboTooltip = "Which t-SNE implementation to run? You can select between FIt-SNE and Barnes-Hut t-SNE. FIt-SNE is the default; Barnes-Hut t-SNE is provided only for comparison.";
    public static final String algOptionName = "alg";
    public static final String[] algOptionsHuman = new String[] { "FIt-SNE", "Barnes-Hut" };
    public static final String[] algOptions = new String[] { "fft", "bh" };
    
    private FJComboBox fKnnAlgCombo = new FJComboBox(knnOptionsHuman);
    public static final String knnAlgComboText = "Nearest neighbors";
    public static final String knnAlgComboTooltip = "Which implementation of nearest neighbors calculation to use to determine the set of points (cells) that each data point (cell) is attracted to? This can be done either by using an approximate nearest neighbors implementation using the Spotify Annoy library (https://github.com/spotify/annoy) or using an exact nearest neighbors implementation with vantage point trees. If the input dimensionality (=number of selected parameters) is relatively small (e.g.., flow cytometry input, or a “few” principal components from single cell RNA seq data) then the exact implementation should be almost as fast as the approximate one, but for high-dimensional (~30+) inputs, the approximate nearest neighbors will perform significantly faster.";
    public static final String knnAlgOptionName = "knn";
    public static final String[] knnOptionsHuman = new String[] { "Approximate", "VP Trees" };
    public static final String[] knnOptions = new String[] { "ann", "vptree" };
    
    private RangedDoubleTextField fPerplexityField = null;
    public static final String perplexityText = "Perplexity";
    public static final String perplexityTooltip = "Perplexity is a measure for information that is defined as 2 to the power of the Shannon entropy. Same as in regular t-SNE, the perplexity may be viewed as a knob that sets the number of effective nearest neighbors that each cell is being attracted by (while also being repulsed by all of the neighbors). In practical terms, if you are getting a strange ‘ball’ with uniformly distributed points, it usually indicates that your perplexity way too high because all points want to be equidistant. If perplexity is set to a negative value then K (kernel width; number of nearest neighbors) and sigma (bandwidth) parameters will be used instead of perplexity.";
    public static final String perplexityOptionName = "perplexity";
    public static final double perplexityMin = -1, perplexityMax = 200, perplexityDefault = 20;
        
    private RangedIntegerTextField fMaxIterField = null;
    public static final String maxIterText = "Max iterations";
    public static final String maxIterTooltip = "The maximum number of iterations that the algorithm will perform.";
    public static final String maxIterOptionName = "maxIter";
    public static final int maxIterMin = 10, maxIterMax = 100000, maxIterDefault = 1000;

    private RangedIntegerTextField fStopLyingIterField = null;
    public static final String stopLyingIterText = "Stop lying at";
    public static final String stopLyingIterTooltip = "Iteration after which the early exaggeration stops and perplexities are no longer exaggerated by the exaggeration factor, i.e., this value controls the length of the early exaggeration phase. Increasing the early exaggeration may be useful for large numbers of events; however, be careful not to increase both the exaggeration and exaggeration factor too much, or nearest neighbors calculations may tend to over-cluster.";
    public static final String stopLyingIterOptionName = "stopLyingIter";
    public static final int stopLyingMin = 10, stopLyingMax = 5000, stopLyingDefault = 250;
    
    private RangedDoubleTextField fExaggerationFactorField = null;
    public static final String exaggerationFactorText = "Exaggeration factor";
    public static final String exaggerationFactorTooltip = "Exaggeration factor used to multiply the attractive forces with during the early exaggeration phase of the calculation. This also controls how tight natural clusters in the original space are in the embedded space and how much space will be between them. If you think about clustered data, the purpose of early exaggeration is to make it so that the cells corresponding to a cluster can all \"find each other\" despite the repulsion from so many other cells.";
    public static final String exaggerationFactorOptionName = "egFact";
    public static final double exaggerationFactorMin = 0.1, exaggerationFactorMax = 100.0, exaggerationFactorDefault = 12.0;
    
    private RangedDoubleTextField fThetaField = null;
    public static final String thetaText = "Theta";
    public static final String thetaTooltip = "Theta controls the estimation of forces for the Barnes Hut algorithm (it is only applicable to the Barnes Hut implementation). Smaller values will lead to more precise (and computationally more expensive implementation); a value of 0 will effectively turn off Barnes Hut and fall back to the original “precise” t-SNE implementation. See also https://jheer.github.io/barnes-hut/ for an explanation of how the choice of theta effects the algorithm.";
    public static final String thetaOptionName = "theta";
    public static final double thetaMin = 0.0, thetaMax = 1.0, thetaDefault = 0.5;

    private RangedDoubleTextField fLateEgCoeffField = null;
    public static final String lateEgCoeffText = "Late exag. coeff";
    public static final String lateEgCoeffTooltip = "Late exaggeration coefficient used to multiply the attractive forces with during the late exaggeration phase, i.e., after the late exaggeration start. Late exaggeration will typically cause the clusters in the resulting embedded space to become tighter and better separable from one another, but you may start losing local structures within those clusters.";
    public static final String lateEgCoeffOptionName = "lateEgCoeff";
    public static final double lateEgCoeffMin = 0.1, lateEgCoeffMax = 100.0, lateEgCoeffDefault = 1.0;

    private RangedIntegerTextField fStartLateExagIterField = null;
    public static final String startLateExagIterText = "Late exag. start";
    public static final String startLateExagIterTooltip = "Iteration at which the late exaggeration starts and attractive forces are again exaggerated, this time by the late exaggeration coefficient. The default value of -1 will disable the late exaggeration.";
    public static final String startLateExagIterOptionName = "lateEgItStart";
    public static final int startLateExagIterMin = -1, startLateExagIterMax = 100000, startLateExagIterDefault = -1;

    private RangedIntegerTextField fNumTreesField = null;
    public static final String numTreesText = "Number of trees";
    public static final String numTreesTooltip = "Number of trees used in the Spotify Annoy library to perform approximate nearest neighbor search. This affects the build time and the size of the index used for nearest neighbor search; a larger value will give more accurate nearest neighbors, but will consume more memory and take more time. See https://github.com/spotify/annoy for more details.";
    public static final String numTreesOptionName = "numTrees";
    public static final int numTreesMin = 2, numTreesMax = 1000, numTreesDefault = 50;

    private RangedIntegerTextField fSearchKField = null;
    public static final String searchKText = "Search K";
    public static final String searchKTooltip = "Search K used in the Spotify Annoy library to perform approximate nearest neighbor search. A value of -1 will set this automatically. This also affects the nearest neighbor search performance; a larger value will give more accurate results, but will take longer time to return. See https://github.com/spotify/annoy for more details.";
    public static final String searchKOptionName = "searchK";
    public static final int searchKMin = -1, searchKMax = 100, searchKDefault = -1;
    
    private RangedIntegerTextField fNumTermsField = null;
    public static final String numTermsText = "Number of terms";
    public static final String numTermsTooltip = "FIt-SNE uses an interpolation scheme to approximate the repulsive forces at each step of gradient descent. This parameter corresponds to the number of interpolation nodes. Increasing the number of nodes should increase the accuracy, but this effect will quickly saturate and the Runge phenomenon (https://en.wikipedia.org/wiki/Runge%27s_phenomenon) takes over because it is an equispaced interpolation. For this reason, it is not recommended to adjust this parameter. Instead, the accuracy of the approximation may be increased by changing the grid using the \"Minimum intervals\" and \"Intervals per integer\".";
    public static final String numTermsOptionName = "numTerms";
    public static final int numTermsMin = 2, numTermsMax = 5, numTermsDefault = 3;

    private RangedIntegerTextField fKField = null;
    public static final String kText = "K";
    public static final String kTooltip = "Number of nearest neighbors to control the kernel width; a value of -1 will auto set this parameter. Note that this parameter will be ignored unless perplexity is set to a negative value, i.e., the kernel width can be controlled either by perplexity, or set manually by K and sigma. Same as perplexity, this may be viewed as a knob that sets the number of effective nearest neighbors that each cell is being attracted by.";
    public static final String kOptionName = "k";
    public static final int kMin = -1, kMax = 100, kDefault = -1;
    
    private RangedDoubleTextField fSigmaField = null;
    public static final String sigmaText = "Sigma";
    public static final String sigmaTooltip = "The bandwidth (standard deviation for Gaussian kernel). This will be ignored unless perplexity is set to a negative value, i.e., the kernel width can be controlled either by perplexity, or set manually by K and sigma. Same as perplexity, this may be viewed as a knob that sets the number of effective nearest neighbors that each cell is being attracted by.";
    public static final String sigmaOptionName = "sigma";
    public static final double sigmaMin = -1000.0, sigmaMax = 1000.0, sigmaDefault = -30.0;
    
    private RangedIntegerTextField fMinNumIntervalsField = null;
    public static final String minNumIntervalsText = "Minimum intervals";
    public static final String minNumIntervalsTooltip = "The minimum number of intervals and the intervals per integer adjust the fineness of the grid used in the FIt-SNE interpolation scheme. The Minimum intervals describes the smallest number of subintervals in which to split each dimension of the t-SNE space when calculating the FIt-SNE. However, as the space increases in size (t-SNE typically expands to be on the order of -100 to 100 by -100 to 100), we need to maintain the fineness of the grid. This is where the Intervals per integer prevail. The number of subintervals in each dimension at any given iteration is the larger of (i) the minimum number of intervals, and (ii) the number of intervals per integer times the current size of the t-SNE grid.";
    public static final String minNumIntervalsOptionName = "minNumIntervals";
    public static final int minNumIntervalsMin = 1, minNumIntervalsMax = 1000, minNumIntervalsDefault = 50;

    private RangedDoubleTextField fIntervalsPerIntegerField = null;
    public static final String intervalsPerIntegerText = "Intervals per integer";
    public static final String intervalsPerIntegerTooltip = "The intervals per integer and the minimum number of intervals adjust the fineness of the grid used in the FIt-SNE interpolation scheme. The Minimum intervals describes the smallest number of subintervals in which to split each dimension of the t-SNE space when calculating the FIt-SNE. However, as the space increases in size (t-SNE typically expands to be on the order of -100 to 100 by -100 to 100), we need to maintain the fineness of the grid. This is where the Intervals per integer prevail. The number of subintervals in each dimension at any given iteration is the larger of (i) the minimum number of intervals, and (ii) the number of intervals per integer times the current size of the t-SNE grid..";
    public static final String intervalsPerIntegerOptionName = "IntPerInt";
    public static final double intervalsPerIntegerMin = 0.1, intervalsPerIntegerMax = 10.0, intervalsPerIntegerDefault = 1.0;

    private RangedIntegerTextField fSeedField = null;
    public static final String seedText = "Seed";
    public static final String seedTooltip = "FItSNE is a stochastic algorithm and as such, slightly different results will be produced every time you run it unless you set seed to a positive value. The default of -1 will set the seed randomly.";
    public static final String seedOptionName = "seed";
    public static final int seedMin = -1, seedMax = 1000000000, seedDefault = -1;
    
    private FJCheckBox fNoMomentumDuringExagCheckbox = null; 
    public static final String noMomentumDuringExagText = "No momentum during exaggeration";
    public static final String noMomentumDuringExagToolTip = "If checked, then there will be no momentum/gains until after the early exaggeration phase is completed, i.e., a regular gradient descent will be used to optimize the cost function. Gradient descent with momentum is an optimization of a gradient descent algorithm that relies on exponentially decreasing weighted averages applied to older data points and thus typically minimizes oscillation of the gradient descent. You can read more at http://datalya.com/blog/2017/gradient-descent-with-momentum but this checkbox should have minimal effect on the results.";
    public static final String noMomentumDuringExagOptionName = "noMomentumDuringExag";
    public static final boolean noMomentumDuringExagDefault = false;
    
    private FJFormattedTextField fDerParNameField = null;
    public static final String derParNameText = "Run ID";
    public static final String derParNameToolTip = "An identifier used to create the derived parameter names. The \"auto\" value will create names encoding most of the input settings, which ensures that they will be unique, but they are also somewhat long due to the number of all the different options provided. You may want to choose your own Run ID to shorten the names.";
    public static final String derParNameOptionName = "derParName";
    public static final String derParNameDefault = "auto";
    
    private RangedIntegerTextField fLearningRateField = null;
    public static final String learningRateText = "Learning rate";
    public static final String learningRateTooltip = "The step size in t-SNE gradient descent is updated at each iteration per Jacobs adaptive learning rate. A value of 200 is a typical t-SNE default. You can also use a value of 0, which will calculate an optimal learning rate based on the input size as per https://doi.org/10.1101/451690.";
    public static final String learningRateOptionName = "learningRate";
    public static final int learningRateMin = 0, learningRateMax = 1000000000, learningRateDefault = 200; // 0 will be auto based on number of events.
    
    public static String makeHTMPar(String s, int w, boolean bold) {
		String ret = "<html><p style=\"width: " + w + "px;\">";
		if (bold) ret += "<b>";
		ret += s;
		if (bold) ret += "</b>";
		ret += "</p></html>";
		return ret;
	}
	
    public FItSNE()
	{
		super(pluginName);
		fShowOutput = true;
	}

	public String getName() 
	{
		return(pluginName);
	}

	public String getVersion() 
	{
		return(pluginVersion);
	}
	
	@Override
	public Icon getIcon()
	{
		if (myIcon == null)
		{
			URL url = getClass().getClassLoader().getResource(getIconName());
			if (url != null)
				myIcon = new ImageIcon(url);
		}
		return myIcon;
	}
	
	@Override
	protected String getIconName()
	{
		return sIconName;
	}
	
	@Override
	protected boolean showClusterInputField()
	{
		return false;
	}
	
	@Override
	public ExportFileTypes useExportFileType() {
		return ExportFileTypes.CSV_CHANNEL;
	}

	public ExternalAlgorithmResults invokeAlgorithm(SElement fcmlQueryElement, File sampleFile, File outputFolder) 
	{
        ExternalAlgorithmResults results = new ExternalAlgorithmResults();
                
        if(!sampleFile.exists()) 
        {
        	JOptionPane.showMessageDialog(null, "Input file did not exist.", getName() + " error", JOptionPane.ERROR_MESSAGE);
            results.setWorkspaceString(FItSNE.Failed);
            return results;
        } 
        else
        {
        	PopNode pN = FJPluginHelper.getParentPopNode(fcmlQueryElement);
        	String pref = "";
        	if (pN != null) {
        		for (String s : pN.getPath()) {
        			s = FilenameUtils.fixFileNamePart(s);
        			pref += s + "_";
        		}
        	}
            String sampleName = StringUtil.rtrim(sampleFile.getName(), ".fcs");
            FItSNERFlowCalc calculator = new FItSNERFlowCalc();
            File algResult = calculator.runAlg(sampleFile, sampleName, preprocessCompParameterNames(), fOptions, outputFolder.getAbsolutePath(), useExistingFiles(), pref);
            if (alreadyExists.contentEquals(algResult.getName()))
            	return results;
            calculator.deleteScriptFile();
            checkROutFile(calculator);
            
			File rErrFile = new File(algResult.getAbsolutePath().toString() + "_err.txt");
			if (rErrFile.exists()) {
				// Something went wrong, there is an error message in the file.
				String errMsg = "Something went wrong";
				try {
					errMsg = new String(Files.readAllBytes(Paths.get(rErrFile.getAbsolutePath())));
				} catch (IOException e) {
					e.printStackTrace();
				}
	        	JOptionPane.showMessageDialog(null, errMsg, getName() + " error", JOptionPane.ERROR_MESSAGE);
	        	results.setWorkspaceString(FItSNE.Failed);
				rErrFile.delete();
			} else {            
	            if (algResult != null && algResult.exists()) {
	        		FJFileRef xyFileRef = null;
	            	try {
	        			xyFileRef = FJFileRefFactory.make(algResult.getAbsolutePath());
	        			if (!isSeqGeq()) {
	        				FJFileRef ref = new RFlowCalculator().generateDerivedParameterCSVFile(xyFileRef, PluginHelper.getNumTotalEvents(fcmlQueryElement), sampleName, outputFolder.getAbsolutePath());
	        				File finalCsvFile = ref.getLocalFile();
	        				PluginHelper.createClusterParameter(results, "plugin", finalCsvFile);
	        			} else {
	        				try {
								boolean result = ((CSVSample)FJPluginHelper.getSample(fcmlQueryElement)).mergeDataFile(FJFileRefFactory.make(algResult.toURI()), false, false);
								if (!result) {
									JOptionPane.showMessageDialog(null, "Failed to merge in " + algResult.getAbsolutePath(), " error", JOptionPane.ERROR_MESSAGE);
									results.setWorkspaceString(FItSNE.Failed);
								}
							} catch (Exception e) {
								e.printStackTrace();
							}        				
	        			}
	        			
	        		} catch (IOException e) {
	        			e.printStackTrace();
	        		}
	            	
	            } else {
	            	JOptionPane.showMessageDialog(null, "Failed to create output.", getName() + " error", JOptionPane.ERROR_MESSAGE);
	                results.setWorkspaceString(FItSNE.Failed);
	            }
			}

            try {
            	if (sampleFile.getParentFile().getName().compareTo(getName()) == 0) {
            		// If the file is in a directory that has the same name as the plugin name then this file was for sure created by FlowJo or SeqGeq
            		// => Let's delete it to make sure it's updated next time the same plugin is executed in case the user selected different parameters
            		// (FlowJo, SeqGeq will sometimes reuse files, which is wrong if different parameters are selected by the user)
            		sampleFile.delete();
            	}
            } catch (Exception e) {};
            
            return results;
        }
	}

	@Override
	protected List<Component> getPromptComponents(SElement selement, List<String> list)
    {
		ArrayList<Component> componentList = new ArrayList<Component>();

		fParameterNameList = new FJList(new DefaultListModel<Object>());
		fParameterNameList.setSelectionMode(2);
		JScrollPane parListScrollPane = new JScrollPane(fParameterNameList);
		GuiFactory.setSizes(parListScrollPane, new Dimension(100, Math.min(200, list.size() * 20)));
		componentList.add(parListScrollPane);
        
        // Set defaults and then retrieve saved options and use them instead of defaults.
        Iterator<SElement> iterator = selement.getChildren("Option").iterator();
        int maxIterSaved = maxIterDefault;
        int startLateExagIterSaved = startLateExagIterDefault; 
        int stopLyingSaved = stopLyingDefault;
        int numTreesSaved = numTreesDefault;
        int numTermsSaved = numTermsDefault;
        int searchKSaved = searchKDefault;
        int kSaved = kDefault;
        int minNumIntervalsSaved = minNumIntervalsDefault;
        int seedSaved = seedDefault;
        int learningRateSaved = learningRateDefault;
        double perplexitySaved = perplexityDefault;
        double exaggerationFactorSaved = exaggerationFactorDefault;
        double thetaSaved = thetaDefault;
        double lateEgCoeffSaved = lateEgCoeffDefault;
        double sigmaSaved = sigmaDefault;
        double intervalsPerIntegerSaved = intervalsPerIntegerDefault;
        boolean noMomentumDuringExagSaved = noMomentumDuringExagDefault;
        String derParNameSaved = derParNameDefault;
        
        int dimsOptionSavedIndex = 1; // 2 dims is the default, which is index number 1 in the array of options
        int algOptionSavedIndex = 0;
        int knnAlgOptionSavedIndex = 0;
        
        while(iterator.hasNext()) {
            SElement option = iterator.next();
            
            String tmpDerParNameSaved = option.getString(derParNameOptionName, ""); 
            if (tmpDerParNameSaved != null && !tmpDerParNameSaved.isEmpty())
            	derParNameSaved = sanitize(tmpDerParNameSaved);

            int tmpMaxIterSaved = option.getInt(maxIterOptionName, maxIterMin - 1);
            if (tmpMaxIterSaved >= maxIterMin && tmpMaxIterSaved <= maxIterMax) 
            	maxIterSaved = tmpMaxIterSaved;
            
            int tmpStopLyingSaved = option.getInt(stopLyingIterOptionName, stopLyingMin - 1);
            if (tmpStopLyingSaved >= stopLyingMin && tmpStopLyingSaved <= stopLyingMax) 
            	stopLyingSaved = tmpStopLyingSaved;
            
            int tmpStartLateExagIterSaved = option.getInt(startLateExagIterOptionName, startLateExagIterMin - 1); 
            if (tmpStartLateExagIterSaved >= startLateExagIterMin && tmpStartLateExagIterSaved <= startLateExagIterMax) 
            	startLateExagIterSaved = tmpStartLateExagIterSaved;

            int tmpNumTreesSaved = option.getInt(numTreesOptionName, numTreesMin - 1); 
            if (tmpNumTreesSaved >= numTreesMin && tmpNumTreesSaved <= numTreesMax) 
            	numTreesSaved = tmpNumTreesSaved;
            
            int tmpNumTermsSaved = option.getInt(numTermsOptionName, numTermsMin - 1); 
            if (tmpNumTermsSaved >= numTermsMin && tmpNumTermsSaved <= numTermsMax) 
            	numTermsSaved = tmpNumTermsSaved;

            int tmpSearchKSaved = option.getInt(searchKOptionName, searchKMin - 1); 
            if (tmpSearchKSaved >= searchKMin && tmpSearchKSaved <= searchKMax) 
            	searchKSaved = tmpSearchKSaved;

            int tmpKSaved = option.getInt(kOptionName, kMin - 1); 
            if (tmpKSaved >= kMin && tmpKSaved <= kMax) 
            	kSaved = tmpKSaved;
            
            int tmpMinNumIntervalsSaved = option.getInt(minNumIntervalsOptionName, minNumIntervalsMin - 1); 
            if (tmpMinNumIntervalsSaved >= minNumIntervalsMin && tmpMinNumIntervalsSaved <= minNumIntervalsMax) 
            	minNumIntervalsSaved = tmpMinNumIntervalsSaved;

            int tmpSeedSaved = option.getInt(seedOptionName, seedMin - 1); 
            if (tmpSeedSaved >= seedMin && tmpSeedSaved <= seedMax) 
            	seedSaved = tmpSeedSaved;
            
            int tmpLearningRateSaved = option.getInt(learningRateOptionName, learningRateMin - 1); 
            if (tmpLearningRateSaved >= learningRateMin && tmpLearningRateSaved <= learningRateMax) 
            	learningRateSaved = tmpLearningRateSaved;
            
            double tmpPerplexitySaved = option.getDouble(perplexityOptionName, Double.NaN);
            if (tmpPerplexitySaved >= perplexityMin && tmpPerplexitySaved <= perplexityMax) 
            	perplexitySaved = tmpPerplexitySaved;

            double tmpexaggerationFactor = option.getDouble(exaggerationFactorOptionName, Double.NaN);
            if (tmpexaggerationFactor >= exaggerationFactorMin && tmpexaggerationFactor <= exaggerationFactorMax) 
            	exaggerationFactorSaved = tmpexaggerationFactor;

            double tmpTheta = option.getDouble(thetaOptionName, Double.NaN);
            if (tmpTheta >= thetaMin && tmpTheta <= thetaMax) 
            	thetaSaved = tmpTheta;
            
            double tmpLateEgCoeff = option.getDouble(lateEgCoeffOptionName, Double.NaN);
            if (tmpLateEgCoeff >= lateEgCoeffMin && tmpLateEgCoeff <= lateEgCoeffMax) 
            	lateEgCoeffSaved = tmpLateEgCoeff;

            double tmpSavedSigma = option.getDouble(sigmaOptionName, Double.NaN);
            if (tmpSavedSigma >= sigmaMin && tmpSavedSigma <= sigmaMax) 
            	sigmaSaved = tmpSavedSigma;

            double tmpIntervalsPerIntegerSaved = option.getDouble(intervalsPerIntegerOptionName, Double.NaN);
            if (tmpIntervalsPerIntegerSaved >= intervalsPerIntegerMin && tmpIntervalsPerIntegerSaved <= intervalsPerIntegerMax) 
            	intervalsPerIntegerSaved = tmpIntervalsPerIntegerSaved;

    		String noMomentumDuringExagStr = option.getAttributeValue(noMomentumDuringExagOptionName);
    		if (noMomentumDuringExagStr != null && !noMomentumDuringExagStr.isEmpty()) {
    			if (One.equals(noMomentumDuringExagStr) || True.equals(noMomentumDuringExagStr))
    				noMomentumDuringExagSaved = true;
    		}
            
            // Dims
            {
				String savedDimsOption = option.getString(dimOptionName);
				int i = 0;
				if (savedDimsOption != null && !savedDimsOption.isEmpty()) for (String s : dimOptions) {
					if (s.equals(savedDimsOption)) {
						dimsOptionSavedIndex = i;
						continue;
					}
					i++;
				}
            }

            // Alg
            {
				String savedAlgOption = option.getString(algOptionName);
				int i = 0;
				if (savedAlgOption != null && !savedAlgOption.isEmpty()) for (String s : algOptions) {
					if (s.equals(savedAlgOption)) {
						algOptionSavedIndex = i;
						continue;
					}
					i++;
				}
            }

            // KNN Alg
            {
				String savedKnnAlgOption = option.getString(knnAlgOptionName);
				int i = 0;
				if (savedKnnAlgOption != null && !savedKnnAlgOption.isEmpty()) for (String s : knnOptions) {
					if (s.equals(savedKnnAlgOption)) {
						knnAlgOptionSavedIndex = i;
						continue;
					}
					i++;
				}
            }
            
        }
        
        FJLabel dimsLabel = new FJLabel(dimsComboText);
		if (dimsOptionSavedIndex >= 0 && dimsOptionSavedIndex < fDimsCombo.getItemCount()) fDimsCombo.setSelectedIndex(dimsOptionSavedIndex);
        fDimsCombo.setToolTipText(makeHTMPar(dimsComboTooltip, labelW / 2, false));
        dimsLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(fDimsCombo, new Dimension(labelW / 4, labelH));
        GuiFactory.setSizes(dimsLabel, new Dimension(labelW / 6, labelH));

        FJLabel algLabel = new FJLabel(algComboText);
		if (algOptionSavedIndex >= 0 && algOptionSavedIndex < fAlgCombo.getItemCount()) fAlgCombo.setSelectedIndex(algOptionSavedIndex);
        fAlgCombo.setToolTipText(makeHTMPar(algComboTooltip, labelW / 2, false));
        algLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(fAlgCombo, new Dimension(labelW / 4, labelH));
        GuiFactory.setSizes(algLabel, new Dimension(labelW / 6, labelH));
        
        FJLabel knnAlgLabel = new FJLabel(knnAlgComboText);
		if (knnAlgOptionSavedIndex >= 0 && knnAlgOptionSavedIndex < fKnnAlgCombo.getItemCount()) fKnnAlgCombo.setSelectedIndex(knnAlgOptionSavedIndex);
        fKnnAlgCombo.setToolTipText(makeHTMPar(knnAlgComboTooltip, labelW / 2, false));
        knnAlgLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(fKnnAlgCombo, new Dimension(labelW / 4, labelH));
        GuiFactory.setSizes(knnAlgLabel, new Dimension(labelW / 6, labelH));

        componentList.add(new HBox(new Component[] { dimsLabel, fDimsCombo, algLabel, fAlgCombo, knnAlgLabel, fKnnAlgCombo }));

        FJLabel perplexityLabel = new FJLabel(perplexityText);
        fPerplexityField = new RangedDoubleTextField(perplexityMin, perplexityMax, new DoubleNumberFormatter());
        fPerplexityField.setDouble(perplexitySaved);
        fPerplexityField.setToolTipText(makeHTMPar(perplexityTooltip, labelW / 2, false));
        fPerplexityField.setHorizontalAlignment(SwingConstants.RIGHT);
        perplexityLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(perplexityLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fPerplexityField, new Dimension(labelW / 4, labelH));
        
        FJLabel maxIterLabel = new FJLabel(maxIterText);
        fMaxIterField = new RangedIntegerTextField(maxIterMin, maxIterMax);
        fMaxIterField.setInt(maxIterSaved);
        fMaxIterField.setToolTipText(makeHTMPar(maxIterTooltip, labelW / 2, false));
        fMaxIterField.setHorizontalAlignment(SwingConstants.RIGHT);
        maxIterLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(maxIterLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fMaxIterField, new Dimension(labelW / 4, labelH));
        
        FJLabel stopLyingIterLabel = new FJLabel(stopLyingIterText);
        fStopLyingIterField = new RangedIntegerTextField(stopLyingMin, stopLyingMax);
        fStopLyingIterField.setInt(stopLyingSaved);
        fStopLyingIterField.setToolTipText(makeHTMPar(stopLyingIterTooltip, labelW / 2, false));
        fStopLyingIterField.setHorizontalAlignment(SwingConstants.RIGHT);
        stopLyingIterLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(stopLyingIterLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fStopLyingIterField, new Dimension(labelW / 4, labelH));
        
        componentList.add(new HBox(new Component[] { perplexityLabel, fPerplexityField, maxIterLabel, fMaxIterField, stopLyingIterLabel, fStopLyingIterField }));

        FJLabel exaggerationFactorLabel = new FJLabel(exaggerationFactorText);
        fExaggerationFactorField = new RangedDoubleTextField(exaggerationFactorMin, exaggerationFactorMax, new DoubleNumberFormatter());
        fExaggerationFactorField.setDouble(exaggerationFactorSaved);
        fExaggerationFactorField.setToolTipText(makeHTMPar(exaggerationFactorTooltip, labelW / 2, false));
        fExaggerationFactorField.setHorizontalAlignment(SwingConstants.RIGHT);
        exaggerationFactorLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(exaggerationFactorLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fExaggerationFactorField, new Dimension(labelW / 4, labelH));

        FJLabel thetaLabel = new FJLabel(thetaText);
        fThetaField = new RangedDoubleTextField(thetaMin, thetaMax, new DoubleNumberFormatter());
        fThetaField.setDouble(thetaSaved);
        fThetaField.setToolTipText(makeHTMPar(thetaTooltip, labelW / 2, false));
        fThetaField.setHorizontalAlignment(SwingConstants.RIGHT);
        thetaLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(thetaLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fThetaField, new Dimension(labelW / 4, labelH));
        
        FJLabel lateEgCoeffLabel = new FJLabel(lateEgCoeffText);
        fLateEgCoeffField = new RangedDoubleTextField(lateEgCoeffMin, lateEgCoeffMax, new DoubleNumberFormatter());
        fLateEgCoeffField.setDouble(lateEgCoeffSaved);
        fLateEgCoeffField.setToolTipText(makeHTMPar(lateEgCoeffTooltip, labelW / 2, false));
        fLateEgCoeffField.setHorizontalAlignment(SwingConstants.RIGHT);
        lateEgCoeffLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(lateEgCoeffLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fLateEgCoeffField, new Dimension(labelW / 4, labelH));
        
        componentList.add(new HBox(new Component[] { exaggerationFactorLabel, fExaggerationFactorField, thetaLabel, fThetaField, lateEgCoeffLabel, fLateEgCoeffField }));
        
        FJLabel startLateExagIterLabel = new FJLabel(startLateExagIterText);
        fStartLateExagIterField = new RangedIntegerTextField(startLateExagIterMin, startLateExagIterMax);
        fStartLateExagIterField.setInt(startLateExagIterSaved);
        fStartLateExagIterField.setToolTipText(makeHTMPar(startLateExagIterTooltip, labelW / 2, false));
        fStartLateExagIterField.setHorizontalAlignment(SwingConstants.RIGHT);
        startLateExagIterLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(startLateExagIterLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fStartLateExagIterField, new Dimension(labelW / 4, labelH));

        FJLabel numTreesLabel = new FJLabel(numTreesText);
        fNumTreesField = new RangedIntegerTextField(numTreesMin, numTreesMax);
        fNumTreesField.setInt(numTreesSaved);
        fNumTreesField.setToolTipText(makeHTMPar(numTreesTooltip, labelW / 2, false));
        fNumTreesField.setHorizontalAlignment(SwingConstants.RIGHT);
        numTreesLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(numTreesLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fNumTreesField, new Dimension(labelW / 4, labelH));

        FJLabel numTermsLabel = new FJLabel(numTermsText);
        fNumTermsField = new RangedIntegerTextField(numTermsMin, numTermsMax);
        fNumTermsField.setInt(numTermsSaved);
        fNumTermsField.setToolTipText(makeHTMPar(numTermsTooltip, labelW / 2, false));
        fNumTermsField.setHorizontalAlignment(SwingConstants.RIGHT);
        numTermsLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(numTermsLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fNumTermsField, new Dimension(labelW / 4, labelH));
        
        componentList.add(new HBox(new Component[] { startLateExagIterLabel, fStartLateExagIterField, numTreesLabel, fNumTreesField, numTermsLabel, fNumTermsField }));

        FJLabel searchKLabel = new FJLabel(searchKText);
        fSearchKField = new RangedIntegerTextField(searchKMin, searchKMax);
        fSearchKField.setInt(searchKSaved);
        fSearchKField.setToolTipText(makeHTMPar(searchKTooltip, labelW / 2, false));
        fSearchKField.setHorizontalAlignment(SwingConstants.RIGHT);
        searchKLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(searchKLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fSearchKField, new Dimension(labelW / 4, labelH));
        
        FJLabel kLabel = new FJLabel(kText);
        fKField = new RangedIntegerTextField(kMin, kMax);
        fKField.setInt(kSaved);
        fKField.setToolTipText(makeHTMPar(kTooltip, labelW / 2, false));
        fKField.setHorizontalAlignment(SwingConstants.RIGHT);
        kLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(kLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fKField, new Dimension(labelW / 4, labelH));
        
        FJLabel sigmaLabel = new FJLabel(sigmaText);
        fSigmaField = new RangedDoubleTextField(sigmaMin, sigmaMax, new DoubleNumberFormatter());
        fSigmaField.setDouble(sigmaSaved);
        fSigmaField.setToolTipText(makeHTMPar(sigmaTooltip, labelW / 2, false));
        fSigmaField.setHorizontalAlignment(SwingConstants.RIGHT);
        sigmaLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(sigmaLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fSigmaField, new Dimension(labelW / 4, labelH));
        
        componentList.add(new HBox(new Component[] { searchKLabel, fSearchKField, kLabel, fKField, sigmaLabel, fSigmaField }));
        
        FJLabel minNumIntervalsLabel = new FJLabel(minNumIntervalsText);
        fMinNumIntervalsField = new RangedIntegerTextField(minNumIntervalsMin, minNumIntervalsMax);
        fMinNumIntervalsField.setInt(minNumIntervalsSaved);
        fMinNumIntervalsField.setToolTipText(makeHTMPar(minNumIntervalsTooltip, labelW / 2, false));
        fMinNumIntervalsField.setHorizontalAlignment(SwingConstants.RIGHT);
        minNumIntervalsLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(minNumIntervalsLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fMinNumIntervalsField, new Dimension(labelW / 4, labelH));

        FJLabel intervalsPerIntegerLabel = new FJLabel(intervalsPerIntegerText);
        fIntervalsPerIntegerField = new RangedDoubleTextField(intervalsPerIntegerMin, intervalsPerIntegerMax, new DoubleNumberFormatter());
        fIntervalsPerIntegerField.setDouble(intervalsPerIntegerSaved);
        fIntervalsPerIntegerField.setToolTipText(makeHTMPar(intervalsPerIntegerTooltip, labelW / 2, false));
        fIntervalsPerIntegerField.setHorizontalAlignment(SwingConstants.RIGHT);
        intervalsPerIntegerLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(intervalsPerIntegerLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fIntervalsPerIntegerField, new Dimension(labelW / 4, labelH));

        FJLabel seedLabel = new FJLabel(seedText);
        fSeedField = new RangedIntegerTextField(seedMin, seedMax);
        fSeedField.setInt(seedSaved);
        fSeedField.setToolTipText(makeHTMPar(seedTooltip, labelW / 2, false));
        fSeedField.setHorizontalAlignment(SwingConstants.RIGHT);
        seedLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(seedLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fSeedField, new Dimension(labelW / 4, labelH));

        componentList.add(new HBox(new Component[] { minNumIntervalsLabel, fMinNumIntervalsField, intervalsPerIntegerLabel, fIntervalsPerIntegerField, seedLabel, fSeedField }));
        
        FJLabel derParNameLabel = new FJLabel(derParNameText);
        fDerParNameField = new FJFormattedTextField();
        fDerParNameField.setText(derParNameSaved);
        fDerParNameField.setToolTipText(makeHTMPar(derParNameToolTip, labelW / 2, false));
        fDerParNameField.setHorizontalAlignment(SwingConstants.RIGHT);
        derParNameLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(derParNameLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fDerParNameField, new Dimension(labelW / 4, labelH));
        
        FJLabel learningRateLabel = new FJLabel(learningRateText);
        fLearningRateField = new RangedIntegerTextField(learningRateMin, learningRateMax);
        fLearningRateField.setInt(learningRateSaved);
        fLearningRateField.setToolTipText(makeHTMPar(learningRateTooltip, labelW / 2, false));
        fLearningRateField.setHorizontalAlignment(SwingConstants.RIGHT);
        learningRateLabel.setHorizontalAlignment(SwingConstants.RIGHT);
        GuiFactory.setSizes(learningRateLabel, new Dimension(labelW / 6, labelH));
        GuiFactory.setSizes(fLearningRateField, new Dimension(labelW / 4, labelH));

        componentList.add(new HBox(new Component[] { derParNameLabel, fDerParNameField, learningRateLabel, fLearningRateField }));
        
        fNoMomentumDuringExagCheckbox = new FJCheckBox(noMomentumDuringExagText);
        fNoMomentumDuringExagCheckbox.setToolTipText(makeHTMPar(noMomentumDuringExagToolTip, labelW / 2, false));
        fNoMomentumDuringExagCheckbox.setSelected(noMomentumDuringExagSaved);
        
        fShowOutputCheckBox = new JCheckBox(keepResultsLabel);
        fShowOutputCheckBox.setToolTipText(makeHTMPar(keepResultsToolTip, labelW / 2, false));
        fShowOutputCheckBox.setSelected(fShowOutput);

        componentList.add(new HBox(new Component[] { fNoMomentumDuringExagCheckbox, fShowOutputCheckBox }));
        
		if (!fRoutFile.isEmpty())
			componentList.add(new HBox(Box.createHorizontalGlue(),  createShowOutputButton(), Box.createHorizontalGlue()));

		FJLabel labelCiting = new FJLabel(makeHTMPar(citingLabel, labelW, false));
        componentList.add(labelCiting);
                
        return componentList;
    }
    
	@SuppressWarnings("deprecation")
	@Override
    protected void extractPromptOptions()
    {
    	fOptions = new HashMap<String, String>();
    	fParameterNames = new ArrayList<String>();
    	boolean cellIdParameterIncluded = false;
    	
    	if (fParameterNameList != null) for (Object obj : fParameterNameList.getSelectedValues()) 
    	{
    		String parName = (new StringBuilder()).append("").append(obj).toString();
    		// FlowJo's parameter names are often in the form of Name :: Description, we only want the Name part from that
    		int parDescIndex = parName.indexOf(" :: ");
    		if(parDescIndex > 0) parName = parName.substring(0, parDescIndex);
    		fParameterNames.add(parName);
    		if (parName.equals(FItSNE.cellIdParName))
    			cellIdParameterIncluded = true;
    	}
    	// We really need the CellId in SeqGeq, so we select it even if the user doesn't.
    	if (isSeqGeq() && !cellIdParameterIncluded)
			fParameterNames.add(FItSNE.cellIdParName);

    	// Save all the plugin specific options
    	fOptions.put(perplexityOptionName, Double.toString(fPerplexityField.getDouble()));
    	fOptions.put(maxIterOptionName, Integer.toString(fMaxIterField.getInt()));
    	fOptions.put(stopLyingIterOptionName, Integer.toString(fStopLyingIterField.getInt()));
    	fOptions.put(exaggerationFactorOptionName, Double.toString(fExaggerationFactorField.getDouble()));
    	fOptions.put(thetaOptionName, Double.toString(fThetaField.getDouble()));
    	fOptions.put(lateEgCoeffOptionName, Double.toString(fLateEgCoeffField.getDouble()));
    	fOptions.put(startLateExagIterOptionName, Integer.toString(fStartLateExagIterField.getInt()));
    	fOptions.put(numTreesOptionName, Integer.toString(fNumTreesField.getInt()));
    	fOptions.put(numTermsOptionName, Integer.toString(fNumTermsField.getInt()));
    	fOptions.put(searchKOptionName, Integer.toString(fSearchKField.getInt()));
    	fOptions.put(kOptionName, Integer.toString(fKField.getInt()));
    	fOptions.put(sigmaOptionName, Double.toString(fSigmaField.getDouble()));
    	fOptions.put(minNumIntervalsOptionName, Integer.toString(fMinNumIntervalsField.getInt()));
    	fOptions.put(intervalsPerIntegerOptionName, Double.toString(fIntervalsPerIntegerField.getDouble()));
    	fOptions.put(seedOptionName, Integer.toString(fSeedField.getInt()));
    	fOptions.put(learningRateOptionName, Integer.toString(fLearningRateField.getInt()));
    	fOptions.put(noMomentumDuringExagOptionName, fNoMomentumDuringExagCheckbox.isSelected() ? One : Zero);
        fOptions.put(derParNameOptionName, sanitize(fDerParNameField.getText()));
    	
    	if (fDimsCombo.getSelectedIndex() < dimOptions.length)
    		fOptions.put(dimOptionName, dimOptions[fDimsCombo.getSelectedIndex()]);
    	else 
    		fOptions.put(dimOptionName, dimOptions[0]);

    	if (fAlgCombo.getSelectedIndex() < algOptions.length)
    		fOptions.put(algOptionName, algOptions[fAlgCombo.getSelectedIndex()]);
    	else 
    		fOptions.put(algOptionName, algOptions[0]);
    	
    	if (fKnnAlgCombo.getSelectedIndex() < knnOptions.length)
    		fOptions.put(knnAlgOptionName, knnOptions[fKnnAlgCombo.getSelectedIndex()]);
    	else 
    		fOptions.put(knnAlgOptionName, knnOptions[0]);
    	
    	fShowOutput = fShowOutputCheckBox.isSelected();
    }

	// Is this SeqGeq based on the application name?
	// This is used to figure out whether we should be including CellIds in the CSV files
	public static boolean isSeqGeq()
	{
        if (FJApplication.getInstance() != null && 
            FJApplication.getInstance().getAppName() != null && 
            FJApplication.getInstance().getAppName().toLowerCase().contains("seqgeq"))
            	return true;
		return false;
	}

	private String sanitize(String text)
	{
		if (text == null || text.isEmpty())
			return "";
		else 
			return text.
					replaceAll(" ", "_").
					replaceAll("<", "_").
					replaceAll(">", "_").
					replaceAll("&", "_").
					replaceAll("\"", "_");
	}
}
